from decimal import Decimal
from typing import Dict, List, Any
from datetime import timedelta, datetime, time, tzinfo

from data_aggregator_sdk.constants.enums import DeviceHack, JournalDataType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, \
    CounterType, ResourceType

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_3phase_consumed
#
# RESULT int:        6653256196879938732392428594653037260
# RESULT bin:  MSB   00000101000000010101111011011011010011000101011101100000000000010000110110010000000000000001111000001111001100000000001011001100   LSB                             # noqa: E501
# RESULT hex:  LE    cc 02 30 0f 1e 00 90 0d 01 60 57 4c db 5e 01 05
#
#
# name                  type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7            7          76                                                                                                                           1001100   # noqa: E501
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1          # noqa: E501
# packet_type_id.1.VAL  u2            2           2                                                                                                                        10           # noqa: E501
# packet_type_id.1.DFF  bool          1           0                                                                                                                       0             # noqa: E501
# energy_is_reactive    bool          1           0                                                                                                                      0              # noqa: E501
# days_ago              timedelta     7           0                                                                                                               0000000               # noqa: E501
# valid                 bool          1           0                                                                                                              0                      # noqa: E501
# total                 u32          32      123123                                                                              00000000000000011110000011110011                       # noqa: E501
# phase_a               u25          25        4313                                                     0000000000001000011011001                                                       # noqa: E501
# phase_b               u25          25    14312123                            0110110100110001010111011
# phase_c               u25          25     1312123   0000101000000010101111011
# RESERVED              u1            1           0  0


class SmpmUlDeviceEnergy16B3PhaseConsumedData(Packet):
    energy_is_reactive: bool
    days_ago: timedelta
    valid: bool
    total: int
    phase_a: int
    phase_b: int
    phase_c: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((76) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((2) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.energy_is_reactive, bool)
        result |= ((int(data.energy_is_reactive)) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 127
        result |= ((days_ago_tmp1) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.valid, bool)
        result |= ((int(data.valid)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.total, int)
        result |= (((data.total) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.phase_a, int)
        result |= (((data.phase_a) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        assert isinstance(data.phase_b, int)
        result |= (((data.phase_b) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        assert isinstance(data.phase_c, int)
        result |= (((data.phase_c) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16B3PhaseConsumedData':
        result__el_tmp2: Dict[str, Any] = dict()
        if 76 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 2 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp2["energy_is_reactive"] = bool(buf.shift(1))
        result__el_tmp2["days_ago"] = timedelta(seconds=buf.shift(7) * 86400)
        result__el_tmp2["valid"] = bool(buf.shift(1))
        result__el_tmp2["total"] = buf.shift(32) + 0
        result__el_tmp2["phase_a"] = buf.shift(25) + 0
        result__el_tmp2["phase_b"] = buf.shift(25) + 0
        result__el_tmp2["phase_c"] = buf.shift(25) + 0
        result = SmpmUlDeviceEnergy16B3PhaseConsumedData(**result__el_tmp2)
        buf.shift(1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        resource_type = ResourceType.ENERGY_REACTIVE if self.energy_is_reactive else ResourceType.ENERGY_ACTIVE
        total_energy_overload_value = 4294967295.0
        hacks = kwargs.get('hacks', [])
        if DeviceHack.electricity_phase_packet_consumed_total_enrg_overload_value in hacks:
            total_energy_overload_value = 3355443.0
        return [
            IntegrationV0MessageData(
                is_valid=self.valid,
                dt=days_ago_calculation(received_at, device_tz, time(0), self.days_ago),
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        value=Decimal(str(self.total)),
                        resource_type=resource_type,
                        channel=1,
                        overloading_value=Decimal(str(total_energy_overload_value)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.ENERGY_PHASE_A,
                        value=Decimal(str(self.phase_a)),
                        resource_type=resource_type,
                        channel=1,
                        overloading_value=Decimal(str(33554431.0)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.ENERGY_PHASE_B,
                        value=Decimal(str(self.phase_b)),
                        resource_type=resource_type,
                        channel=1,
                        overloading_value=Decimal(str(33554431.0)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.ENERGY_PHASE_C,
                        value=Decimal(str(self.phase_c)),
                        resource_type=resource_type,
                        channel=1,
                        overloading_value=Decimal(str(33554431.0)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
            ),
        ]



from typing import Dict, Any, Tuple, List
from datetime import datetime, tzinfo

from data_aggregator_sdk.integration_message import IntegrationV0MessageData

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_ul_device_internal_info
#
# RESULT int:        9187201950435737471
# RESULT bin:  MSB   01111111 01111111 01111111 01111111 01111111 01111111 01111111 01111111   LSB
# RESULT hex:  LE    7f 7f 7f 7f 7f 7f 7f 7f
#
# name                  type  size  value(int)                                                        data(bits)
# --------------------  ----  ----  ----------  ----------------------------------------------------------------
# packet_type_id.0.VAL  u7       7         127                                                           1111111
# packet_type_id.0.DFF  bool     1           0                                                          0
# data.0.byte           u8       8         127                                                  01111111
# data.1.byte           u8       8         127                                          01111111
# data.2.byte           u8       8         127                                  01111111
# data.3.byte           u8       8         127                          01111111
# data.4.byte           u8       8         127                  01111111
# data.5.byte           u8       8         127          01111111
# data.6.byte           u8       8         127  01111111


class SmpmUlDeviceInternalInfoData(Packet):
    data: Tuple[int, int, int, int, int, int, int]

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((127) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.data, tuple) and len(data.data) == 7
        assert isinstance(data.data[0], int)
        assert 0 <= data.data[0] <= 255
        result |= ((data.data[0]) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.data[1], int)
        assert 0 <= data.data[1] <= 255
        result |= ((data.data[1]) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.data[2], int)
        assert 0 <= data.data[2] <= 255
        result |= ((data.data[2]) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.data[3], int)
        assert 0 <= data.data[3] <= 255
        result |= ((data.data[3]) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.data[4], int)
        assert 0 <= data.data[4] <= 255
        result |= ((data.data[4]) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.data[5], int)
        assert 0 <= data.data[5] <= 255
        result |= ((data.data[5]) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.data[6], int)
        assert 0 <= data.data[6] <= 255
        result |= ((data.data[6]) & (2 ** (8) - 1)) << size
        size += 8
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceInternalInfoData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 127 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        data_tmp2: List[int] = []
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        data__item_tmp3 = buf.shift(8) + 0
        data_tmp2.append(data__item_tmp3)
        result__el_tmp1["data"] = tuple(data_tmp2)
        result = SmpmUlDeviceInternalInfoData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(dt=received_at),
        ]



from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_16b_set_regular_data_sending import SmpmDlDeviceEnergy16BSetRegularDataSendingData, \
    SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId, SpecUpLinkModelingPacketIdU8Action
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_dl_device_energy_16b_set_regular_data_sending() -> None:
    case_serialized = bytes.fromhex("96d909f8000000000000000000000000")
    assert SmpmDlDeviceEnergy16BSetRegularDataSendingData(pack_id=SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId.UL_DATA_16B__ENERGY, action=SpecUpLinkModelingPacketIdU8Action.REPLACE, period_ago=31, week_mask=(False, False, False, False, False, False, False), days_mask=(False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False), hour_mask=(False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)) == SmpmDlDeviceEnergy16BSetRegularDataSendingData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("96d90906000000000000000000000000")  # overflow
    assert SmpmDlDeviceEnergy16BSetRegularDataSendingData(pack_id=SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId.UL_DATA_16B__ENERGY, action=SpecUpLinkModelingPacketIdU8Action.REPLACE_DAY, period_ago=0, week_mask=(False, False, False, False, False, False, False), days_mask=(False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False), hour_mask=(False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)) == SmpmDlDeviceEnergy16BSetRegularDataSendingData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("96d909fe010000000000000000000000")  # overflow
    assert SmpmDlDeviceEnergy16BSetRegularDataSendingData(pack_id=SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId.UL_DATA_16B__ENERGY, action=SpecUpLinkModelingPacketIdU8Action.REPLACE_DAY, period_ago=63, week_mask=(False, False, False, False, False, False, False), days_mask=(False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False), hour_mask=(False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)) == SmpmDlDeviceEnergy16BSetRegularDataSendingData.parse(BufRef(case_serialized))  # noqa: E501



import typing

import math


class BufRef:
    def __init__(self, data: typing.Union[bytes, bytearray, int], starts_at: int = 0, stop_on_buffer_end: bool = False) -> None:
        self.value: typing.Union[bytes, bytearray] = data if not isinstance(data, int) else data.to_bytes(math.ceil(max(data.bit_length(), 1) / 8), "little")
        self._stop_on_buffer_end, self.ends_at = stop_on_buffer_end, starts_at

    def get_bits(self, *, size: int, start: int = 0) -> int:
        bytes_i, bit_i = (self.ends_at + start) // 8, (self.ends_at + start) % 8
        res_size, res, needable_len = 0, 0, bytes_i + math.ceil((bit_i + size) / 8)
        buf_len = len(self.value)
        for i in range(bytes_i, needable_len):
            if i >= buf_len:
                if self._stop_on_buffer_end:
                    break
                raise ValueError(f'buffer has no enough elements. {buf_len} < {needable_len}')
            s = min(8 - bit_i, size - res_size)
            res |= ((self.value[i] >> bit_i) & (2 ** s - 1)) << res_size
            res_size += s
            bit_i = 0
        return res

    def shift(self, size: int) -> int:
        bytes_i, bit_i = self.ends_at // 8, self.ends_at % 8
        res_size, res, needable_len = 0, 0, bytes_i + math.ceil((bit_i + size) / 8)
        buf_len = len(self.value)
        for i in range(bytes_i, needable_len):
            if i >= buf_len:
                if self._stop_on_buffer_end:
                    break
                raise ValueError(f'buffer has no enough elements. {buf_len} < {needable_len}')
            s = min(8 - bit_i, size - res_size)
            res |= ((self.value[i] >> bit_i) & (2 ** s - 1)) << res_size
            res_size += s
            bit_i = 0
        self.ends_at += res_size
        return res

    def get_last_bytes(self) -> bytes:
        size = len(self.value) * 8 - self.ends_at
        return self.get_bits(size=size).to_bytes(length=math.ceil(size / 8), byteorder="little")

    def offset(self, size: int) -> None:
        assert size >= 0, f'{size} was given'
        self.ends_at += size



from datetime import datetime, tzinfo
from decimal import Decimal
from typing import List, Any, Dict

from data_aggregator_sdk.constants.enums import SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageEvent, IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_ul_device_gaz_flow_08b_daily
#
# RESULT int:        9423942916432996
# RESULT bin:  MSB   00000000 00100001 01111011 00000110 10110001 11101010 11100100 01100100   LSB
# RESULT hex:  LE    64 e4 ea b1 06 7b 21 00
#
#
# name                            type    size  value(int)                                                        data(bits)
# --------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL            u7         7         100                                                           1100100
# packet_type_id.0.DFF            bool       1           0                                                          0
# cumulative_volume               uf32p3    32   112323300                          00000110101100011110101011100100
# temperature                     u8         8         123                  01111011
# battery_volts                   uf7p1      7          33           0100001
# event_reset                     bool       1           0          0
# event_case_was_opened           bool       1           0         0
# event_flow_reverse              bool       1           0        0
# event_flow_speed_is_over_limit  bool       1           0       0
# event_sensor_error_measurement  bool       1           0      0
# event_sensor_error_temperature  bool       1           0     0
# event_low_battery_level         bool       1           0    0
# event_system_error              bool       1           0   0
# RESERVED                        u1         1           0  0


class SmpmUlDeviceGazFlow08BDailyData(Packet):
    cumulative_volume: float
    temperature: int
    battery_volts: float
    event_reset: bool
    event_case_was_opened: bool
    event_flow_reverse: bool
    event_flow_speed_is_over_limit: bool
    event_sensor_error_measurement: bool
    event_sensor_error_temperature: bool
    event_low_battery_level: bool
    event_system_error: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((100) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.cumulative_volume, (int, float))
        result |= ((int(round(float(data.cumulative_volume) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.temperature, int)
        assert -100 <= data.temperature <= 155
        result |= (((data.temperature + 100)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 12.7
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_case_was_opened, bool)
        result |= ((int(data.event_case_was_opened)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_reverse, bool)
        result |= ((int(data.event_flow_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_speed_is_over_limit, bool)
        result |= ((int(data.event_flow_speed_is_over_limit)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error_measurement, bool)
        result |= ((int(data.event_sensor_error_measurement)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error_temperature, bool)
        result |= ((int(data.event_sensor_error_temperature)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_system_error, bool)
        result |= ((int(data.event_system_error)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceGazFlow08BDailyData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 100 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["cumulative_volume"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp1["temperature"] = buf.shift(8) + -100
        result__el_tmp1["battery_volts"] = round(buf.shift(7) / 10.0, 1)
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_case_was_opened"] = bool(buf.shift(1))
        result__el_tmp1["event_flow_reverse"] = bool(buf.shift(1))
        result__el_tmp1["event_flow_speed_is_over_limit"] = bool(buf.shift(1))
        result__el_tmp1["event_sensor_error_measurement"] = bool(buf.shift(1))
        result__el_tmp1["event_sensor_error_temperature"] = bool(buf.shift(1))
        result__el_tmp1["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp1["event_system_error"] = bool(buf.shift(1))
        result = SmpmUlDeviceGazFlow08BDailyData(**result__el_tmp1)
        buf.shift(1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.cumulative_volume)),
                        overloading_value=Decimal(str(4294967.295)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.CASE_WAS_OPENED] if self.event_case_was_opened else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_sensor_error_temperature else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.FLOW_REVERSE] if self.event_flow_reverse else []),
                    *([IntegrationV0MessageEvent.FLOW_SPEED_OVER_LIMIT] if self.event_flow_speed_is_over_limit else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR_MEASUREMENT] if self.event_sensor_error_measurement else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR_TEMPERATURE] if self.event_system_error else []),
                ],
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_12b_counter import SmpmUlDeviceJupiter12BCounterData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_jupiter_12b_counter() -> None:
    case_serialized = bytes.fromhex("d4f9ffffffefffffffbf5037")
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=8589934591, value_channel_2=8589934591, battery_volts=3.3, temperature=23, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(value_channel_1=8589934591, value_channel_2=8589934591, battery_volts=3.3, temperature=23, event_reset=True, event_low_battery_level=True))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000000e03f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=0, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000000e03f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=0, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000000e03f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=34359738368, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000000e03f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=34359738368, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000000030")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=0, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000000030")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=0, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000000030")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=34359738368, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000000030")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=0.0, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=34359738368, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000080ff3f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=0, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000080ff3f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=0, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000080ff3f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=34359738368, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d4010000000000000080ff3f")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=92, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=92, value_channel_1=34359738368, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000801f30")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=0, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000801f30")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=0, value_channel_2=34359738368))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000801f30")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=34359738368, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("d40100000000000000801f30")  # overflow
    assert SmpmUlDeviceJupiter12BCounterData(value_channel_1=0, value_channel_2=0, battery_volts=6.3, temperature=-35, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter12BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterData.serialize(SmpmUlDeviceJupiter12BCounterData(battery_volts=6.3, event_low_battery_level=True, event_reset=True, temperature=-35, value_channel_1=34359738368, value_channel_2=34359738368))  # noqa: E501



from data_gateway_sdk.protocols.smpm.smpm_ul_device_internal_info import SmpmUlDeviceInternalInfoData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_internal_info() -> None:
    case_serialized = bytes.fromhex("7f7f7f7f7f7f7f7f")
    assert SmpmUlDeviceInternalInfoData(data=(127, 127, 127, 127, 127, 127, 127)) == SmpmUlDeviceInternalInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceInternalInfoData.serialize(SmpmUlDeviceInternalInfoData(data=(127, 127, 127, 127, 127, 127, 127)))  # noqa: E501
    case_serialized = bytes.fromhex("7f7f7f7f7f7f7f7f")  # overflow
    assert SmpmUlDeviceInternalInfoData(data=(127, 127, 127, 127, 127, 127, 127)) == SmpmUlDeviceInternalInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceInternalInfoData.serialize(SmpmUlDeviceInternalInfoData(data=(127, 127, 127, 127, 127, 127, 127)))  # noqa: E501



from datetime import datetime, tzinfo
from typing import Dict, Any, List

from data_aggregator_sdk.integration_message import IntegrationV0MessageData

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_device_info
#
# RESULT int:        1228249668322633963475565502
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000011 11110111 11111011 11111011 11101111 11111111 11111111 11111111 11111111 11111111 11111011 10111110   LSB
# RESULT hex:  LE    be fb ff ff ff ff ff ef fb fb f7 03 00 00 00 00
#
# name                  type  size       value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ----  ----  ---------------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7       7               62                                                                                                                           0111110   # noqa: E501
# packet_type_id.0.DFF  bool     1                1                                                                                                                          1
# packet_type_id.1.VAL  u2       2                3                                                                                                                        11
# packet_type_id.1.DFF  bool     1                0                                                                                                                       0
# manufacturer_number   u50     50  562949953421311                                                                     01111111111111111111111111111111111111111111111111
# device_type           u6       6               31                                                               011111
# firmware_version      u8       8              127                                                       01111111
# ktt                   u9       9              255                                              011111111
# ktn                   u7       7               63                                       0111111
# RESERVED              u37     37                0  0000000000000000000000000000000000000


class SmpmUlDeviceEnergy16BDeviceInfoData(Packet):
    manufacturer_number: int
    device_type: int
    firmware_version: int
    ktt: int
    ktn: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((62) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((3) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.manufacturer_number, int)
        assert 0 <= data.manufacturer_number <= 1125899906842623
        result |= ((data.manufacturer_number) & (2 ** (50) - 1)) << size
        size += 50
        assert isinstance(data.device_type, int)
        assert 0 <= data.device_type <= 63
        result |= ((data.device_type) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.firmware_version, int)
        assert 0 <= data.firmware_version <= 255
        result |= ((data.firmware_version) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.ktt, int)
        assert 0 <= data.ktt <= 511
        result |= ((data.ktt) & (2 ** (9) - 1)) << size
        size += 9
        assert isinstance(data.ktn, int)
        assert 0 <= data.ktn <= 127
        result |= ((data.ktn) & (2 ** (7) - 1)) << size
        size += 7
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BDeviceInfoData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 62 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 3 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["manufacturer_number"] = buf.shift(50) + 0
        result__el_tmp1["device_type"] = buf.shift(6) + 0
        result__el_tmp1["firmware_version"] = buf.shift(8) + 0
        result__el_tmp1["ktt"] = buf.shift(9) + 0
        result__el_tmp1["ktn"] = buf.shift(7) + 0
        result = SmpmUlDeviceEnergy16BDeviceInfoData(**result__el_tmp1)
        buf.shift(37)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                # TODO: extend integration message with new structure for device info and fill it out here
            ),
        ]



import unittest

from data_gateway_sdk.crypto_algorithms.crypto_aes import aes_ecb_decryption, aes_ecb_encryption, AesError
from data_gateway_sdk.crypto_algorithms.crypto_kuznechik import kuznechik_encryption, kuznechik_decryption, KuznechikError


class TestCrypto(unittest.TestCase):
    def test_aes_128(self) -> None:
        # ECB-AES-128 test
        plaintext = bytes.fromhex('6bc1bee22e409f96e93d7e117393172a'
                                  'ae2d8a571e03ac9c9eb76fac45af8e51'
                                  '30c81c46a35ce411e5fbc1191a0a52ef'
                                  'f69f2445df4f9b17ad2b417be66c3710')

        key = bytes.fromhex('2b7e151628aed2a6abf7158809cf4f3c')

        expected_ciphertext = bytes.fromhex('3ad77bb40d7a3660a89ecaf32466ef97'
                                            'f5d3d58503b9699de785895a96fdbaaf'
                                            '43b1cd7f598ece23881b00e3ed030688'
                                            '7b0c785e27e8ad3f8223207104725dd4')

        ciphertext = aes_ecb_encryption(plaintext, key)
        self.assertEqual(ciphertext, expected_ciphertext)

        recovered_plaintext = aes_ecb_decryption(ciphertext, key)
        self.assertEqual(recovered_plaintext, plaintext)

    def test_aes_256(self) -> None:
        # ECB-AES-256 test
        plaintext = bytes.fromhex('00112233445566778899aabbccddeeff'
                                  '00112233445566778899aabbccddeeff'
                                  '00112233445566778899aabbccddeeff'
                                  '00112233445566778899aabbccddeeff')

        key = bytes.fromhex('000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f')

        expected_ciphertext = bytes.fromhex('8ea2b7ca516745bfeafc49904b496089'
                                            '8ea2b7ca516745bfeafc49904b496089'
                                            '8ea2b7ca516745bfeafc49904b496089'
                                            '8ea2b7ca516745bfeafc49904b496089')

        ciphertext = aes_ecb_encryption(plaintext, key)
        self.assertEqual(ciphertext, expected_ciphertext)

        recovered_plaintext = aes_ecb_decryption(ciphertext, key)
        self.assertEqual(recovered_plaintext, plaintext)

        with self.assertRaises(AesError):
            plaintext = bytes.fromhex('00112233445566778899aabbccddeeff'
                                      '00112233445566778899aabbccddeeff'
                                      '00112233445566778899aabbccddeeff'
                                      '00112233445566778899aabbccddeeff')
            key = bytes.fromhex('000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e')
            aes_ecb_encryption(plaintext, key)

    def test_kuznechik(self) -> None:
        mtest = bytes.fromhex('1122334455667700ffeeddccbbaa9988')
        ktest = bytes.fromhex('8899aabbccddeeff0011223344556677fedcba98765432100123456789abcdef')

        encrypted_block = kuznechik_encryption(mtest, ktest)
        expected_enc = bytes.fromhex('7f679d90bebc24305a468d42b9d4edcd')
        self.assertEqual(encrypted_block, expected_enc)

        decrypted_block = kuznechik_decryption(encrypted_block, ktest)
        expected_dec = bytes.fromhex('1122334455667700ffeeddccbbaa9988')
        self.assertEqual(decrypted_block, expected_dec)

        with self.assertRaises(KuznechikError):
            mtest = bytes.fromhex('1122334455667700ffeeddccbbaa9988')
            ktest = bytes.fromhex('8899aabbccddeeff0011223344556677fedcba98765432100123456789abcd')
            kuznechik_encryption(mtest, ktest)

        with self.assertRaises(KuznechikError):
            mtest = bytes.fromhex('1122334455667700ffeeddccbbaa99')
            ktest = bytes.fromhex('8899aabbccddeeff0011223344556677fedcba98765432100123456789abcdef')
            kuznechik_encryption(mtest, ktest)


if __name__ == "__main__":
    unittest.main()



from datetime import datetime, timedelta

from data_gateway_sdk.device_data_protocol import DeviceProtocolType
from data_gateway_sdk.protocols.device_packet.ncp_smp_device_packet import NcpSmpV0DevicePacket
from data_gateway_sdk.protocols.device_packet.smp_device_packet import SmpV0DevicePacket, SmpDaily
from data_gateway_sdk.protocols.device_packet.smpm_device_packet import SmpMGasMeterV0DevicePacket, \
    SmpMEnergyMeterV0DevicePacket, SmpMJupiter08BV0DevicePacket, \
    SmpMJupiter12BV0DevicePacket, SmpMJupiter16BV0DevicePacket, SmpMWaterMeter04BV0DevicePacket, \
    SmpMWaterMeter08BV0DevicePacket, SmpMWaterMeter12BV0DevicePacket, \
    SmpMWaterMeter16BV0DevicePacket
from data_gateway_sdk.protocols.device_packet.water5_device_packet import Water5NeroV0DevicePacket
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_3phase_consumed import \
    SmpmUlDeviceEnergy16B3PhaseConsumedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_3phase_generated import \
    SmpmUlDeviceEnergy16B3PhaseGeneratedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_daily import SmpmUlDeviceEnergy16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_info import SmpmUlDeviceEnergy16BInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_journal import SmpmUlDeviceEnergy16BJournalData, \
    SmpmUlDeviceEnergy16bJournalType, EventData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h3_energy import \
    SmpmUlDeviceEnergy16bProfile8h3EnergyType, SmpmUlDeviceEnergy16BProfile8H3EnergyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_tariff_consumed import \
    SmpmUlDeviceEnergy16BTariffConsumedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_tariff_generated import \
    SmpmUlDeviceEnergy16BTariffGeneratedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_08b_daily import SmpmUlDeviceGazFlow08BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_16b_daily import SmpmUlDeviceGazFlow16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_32b_daily import SmpmUlDeviceGazFlow32BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter import SmpmUlDeviceJupiter08BCounterData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_ch1 import SmpmUlDeviceJupiter08BCounterCh1Data
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_ch2 import SmpmUlDeviceJupiter08BCounterCh2Data
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_12b_counter import SmpmUlDeviceJupiter12BCounterData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_16b_counter import SmpmUlDeviceJupiter16BCounterData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_04b_event import SmpmUlDeviceWaterMeter04BEventData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_daily import SmpmUlDeviceWaterMeter08BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_daily_new import \
    SmpmUlDeviceWaterMeter08BDailyNewData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_info import SmpmUlDeviceWaterMeter08BInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_12b_daily import SmpmUlDeviceWaterMeter12BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_16b_daily import SmpmUlDeviceWaterMeter16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_nbiot_16b_daily import \
    SmpmUlDeviceWaterMeterNbiot16BDailyData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_cold_reset import Water5DeviceColdResetData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_daily import Water5DeviceDailyData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_info_ch1 import Water5DeviceInfoCh1Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_magnet import Water5DeviceMagnetData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch1 import Water5DeviceWeeklyCh1Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch2 import Water5DeviceWeeklyCh2Data


# FLUO(FREESCALE) , NBFI
# assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('97157A6F8028CF075DE4DC4EB9752171EBAF87E704E3C12E7905126ED48747946068D9E4'))


# FLUO A, NBFI
# assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('97157A6F8028CF075DE4DC4EB9752171EBAF87E704E3C12E7905126ED48747946068D9E4'))


def test_protocol_water5_v_nero_v0() -> None:
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('54c2000000000000')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceDailyData(value=24.874),
    )]

    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('73086e0a000000a5')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=683.528,
            # battery=3.37,
            battery_int=3,
            battery_fract=37,
        ),
    )]

    assert 3 + (37 / 100) == 3.37

    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('7b3a00000000001f')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh2Data(
            value=0.058,
            # battery=2.31,
            battery_int=2,
            battery_fract=31,
        ),
    )]

    # FLUO A, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000 ПОТОМУ ЧТО м3 а не в литрах
            # battery=3.67,
            battery_int=3,
            battery_fract=67,
        ),
    )]

    # FLUO A, W5, INFO
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC4894470000D9')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда, В ДОКЕ УКАЗАНО 00 !!!!!(возможно опечатка)
        ),
    )]

    # FLUO A, W5, RESET
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('41CC489447014100')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceColdResetData(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
    )]

    # FLUO A, W5, WARM_RESET
    # assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('41CC4894470100')) == Water5NeroV0DeviceData(
    #     cold_reset=Water5WarmReset(
    #         ???
    #     ),
    # )

    # FLUO(FREESCALE), W5, DAILY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('6086000000000000')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceDailyData(
            value=17200 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
    )]

    # FLUO(FREESCALE), W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
    )]

    # FLUO(FREESCALE), W5, WEEKLY (исп.0), Размерность переменной накопленного объема в прямом направлении за все время равна 24 бит вместо 27 бит.
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('73DA45A3000000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=10700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            # battery=3.67,
            battery_int=3,
            battery_fract=67,
        ),  # counter_bit_size исходя из описания возможно максимум 24???
    )]
    # WEEKLY и WEEKLY (исп.0) дают идентичный результат

    # FLUO(FREESCALE), W5, INFO, Размерность переменной накопленного объема в прямом направлении за все время равна 24 бит вместо 27 бит.
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC4894470000D9')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
    )]

    # FLUO(FREESCALE), W5, INFO(исп.0), Присутствует значение переменной общего количества отправленных пакетов count_send_message
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC489447F875D9')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
    )]

    print(1111111111111111111111111111111)
    # # FLUO(FREESCALE), W5, RESET
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('41CC489447014100')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceColdResetData(
            value=1200900300 / 1000,
        ),
    )]

    # FLUO(FREESCALE), W5, COMMAND
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('51AAAAAAAAAAAAAA')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceMagnetData(),
    )]

    # FLUO(FREESCALE), W5, WARM_RESET
    # assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('43820000004E4552')) == Water5NeroV0DeviceData(
    #     command=Water5WarmReset(
    #         ???
    #     ),
    # )

    # JUPITER FREESCALE/STM32, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31000000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=3259738 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            # battery=3.67,
            battery_int=3,
            battery_fract=67,
        ),
    )]

    # JUPITER FREESCALE/STM32, W5, WARM_RESET
    # assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('43210000004E4552')) == Water5NeroV0DeviceData(
    #     weekly=Water5WarmReset(
    #         ???
    #     ),
    # )

    # FLUO LoRa, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            # battery=3.67,
            battery_int=3,
            battery_fract=67,
        ),
    )]

    # FLUO STM, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            # battery=3.67,
            battery_int=3,
            battery_fract=67,
        ),
    )]
    # FLUO LoRa, W5, WEEKLY = FLUO STM, W5, WEEKLY

    # FLUO STM, W5, INFO
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC4894470000D9')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
    )]

    # FLUO STM, W5, COMMAND
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('51AAAAAAAAAAAAAA')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceMagnetData(),
    )]

    # FLUO STM, W5, WARM_RESET
    # assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('43000000104E4552')) == Water5NeroV0DeviceData(
    #     ??
    # )

    # Jupiter LoRa, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3')) == [Water5NeroV0DevicePacket(  # type: ignore
        packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            # battery=3.67,
            battery_int=3,
            battery_fract=67,
        ),
    )]


def test_protocol_smp_v0() -> None:
    assert DeviceProtocolType.SMP_V0.parse(bytes.fromhex("09800605030000016ed88d016fd999051f1620dc02318b9a9ffc05")) == [SmpV0DevicePacket(
        packet=SmpDaily(
            info_id_code=3,
            sdata_count_ch1_code=18136,
            sdata_count_ch2_code=85209,
            sdata_temperature_code=22,
            sdata_battery_code=3.48,
            sdata_datetime_code=datetime(year=2062, month=10, day=15, hour=4, minute=16, second=11),
        ),
    )]

    # Jupiter NBIoT, SMP, daily
    assert DeviceProtocolType.SMP_V0.parse(bytes.fromhex("09800605AC020000016ED4576F98431F1620EF0231F893CB8306")) == [SmpV0DevicePacket(
        packet=SmpDaily(
            info_id_code=300,
            sdata_count_ch1_code=11220,
            sdata_count_ch2_code=8600,
            sdata_temperature_code=22,
            sdata_battery_code=367 / 100,  # по документации
            sdata_datetime_code=datetime(2063, 4, 11, 10, 5, 44),
        ),
    )]

    # Metano NBIoT, SMP, daily
    assert DeviceProtocolType.SMP_V0.parse(bytes.fromhex("09800605AC020000016ED4576F98431F1620EF0231F893CB8306")) == [SmpV0DevicePacket(
        packet=SmpDaily(
            info_id_code=300,
            sdata_count_ch1_code=11220,
            sdata_count_ch2_code=8600,
            sdata_temperature_code=22,
            sdata_battery_code=367 / 100,  # по документации
            sdata_datetime_code=datetime(2063, 4, 11, 10, 5, 44),
        ),
    )]
    # Jupiter NBIoT, SMP, daily = Metano NBIoT, SMP, daily


def test_protocol_smpm_gaz_v0() -> None:
    assert DeviceProtocolType.SMP_M_GAS_METER_V0.parse(bytes.fromhex('64e4eab1067b2100')) == [SmpMGasMeterV0DevicePacket(  # type: ignore
        packet=SmpmUlDeviceGazFlow08BDailyData(
            cumulative_volume=112323.3,
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_case_was_opened=False,
            event_flow_reverse=False,
            event_flow_speed_is_over_limit=False,
            event_sensor_error_measurement=False,
            event_sensor_error_temperature=False,
            event_low_battery_level=False,
            event_system_error=False,
        ),
    )]


def test_protocol_ncp_smp_v0() -> None:
    assert DeviceProtocolType.NCP_SMP_V0.parse(bytes.fromhex("0603000000000609800605030000016ed88d016fd999051f1620dc02318b9a9ffc052df2")) == [NcpSmpV0DevicePacket(
        mac=3,
        packet=SmpDaily(
            info_id_code=3,
            sdata_count_ch1_code=18136,
            sdata_count_ch2_code=85209,
            sdata_temperature_code=22,
            sdata_battery_code=3.48,
            sdata_datetime_code=datetime(year=2062, month=10, day=15, hour=4, minute=16, second=11),
        ),
    )]


def test_protocol_smpm_energy_v0_multiple_parsing() -> None:
    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("cc02300f1e00900d0160574cdb5e0105cc02300f1e00900d0160574cdb5e0105")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(
            packet=SmpmUlDeviceEnergy16B3PhaseConsumedData(
                energy_is_reactive=False,
                days_ago=timedelta(seconds=0.0),
                valid=False, total=123123,
                phase_a=4313,
                phase_b=14312123,
                phase_c=1312123,
            )),
        SmpMEnergyMeterV0DevicePacket(
            packet=SmpmUlDeviceEnergy16B3PhaseConsumedData(
                energy_is_reactive=False,
                days_ago=timedelta(seconds=0.0),
                valid=False, total=123123,
                phase_a=4313,
                phase_b=14312123,
                phase_c=1312123,
            )),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("cb02300f1e00900d0160574cdb5e0105cb02300f1e00900d0160574cdb5e0105")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(
            packet=SmpmUlDeviceEnergy16B3PhaseGeneratedData(
                energy_is_reactive=False,
                days_ago=timedelta(seconds=0.0),
                valid=False,
                total=123123,
                phase_a=4313,
                phase_b=14312123,
                phase_c=1312123,
            ),
        ),
        SmpMEnergyMeterV0DevicePacket(
            packet=SmpmUlDeviceEnergy16B3PhaseGeneratedData(
                energy_is_reactive=False,
                days_ago=timedelta(seconds=0.0),
                valid=False,
                total=123123,
                phase_a=4313,
                phase_b=14312123,
                phase_c=1312123,
            ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("bbaaaaae545f55aaaa2abbaa6a80160bbbaaaaae545f55aaaa2abbaa6a80160b")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(
            packet=SmpmUlDeviceEnergy16BDailyData(
                energy_consumed_active=1430869,
                energy_consumed_reactive=1398741,
                energy_generated_active=1398101,
                energy_generated_reactive=6990523,
                days_ago=timedelta(seconds=0.0),
                valid=False,
                error_measurement=True,
                error_low_voltage=False,
                error_internal_clock=True,
                error_flash=True,
                error_eeprom=False,
                error_radio=True,
                error_display=False,
                error_plc=False,
                error_reset=False,
                impact_power_lost=True,
                impact_magnet=True,
                impact_cleat_tamper=False,
                impact_body_tamper=True,
                impact_radio=False,
            ),
        ),
        SmpMEnergyMeterV0DevicePacket(
            packet=SmpmUlDeviceEnergy16BDailyData(
                energy_consumed_active=1430869,
                energy_consumed_reactive=1398741,
                energy_generated_active=1398101,
                energy_generated_reactive=6990523,
                days_ago=timedelta(seconds=0.0),
                valid=False,
                error_measurement=True,
                error_low_voltage=False,
                error_internal_clock=True,
                error_flash=True,
                error_eeprom=False,
                error_radio=True,
                error_display=False,
                error_plc=False,
                error_reset=False,
                impact_power_lost=True,
                impact_magnet=True,
                impact_cleat_tamper=False,
                impact_body_tamper=True,
                impact_radio=False,
            ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("bc0aed017c5687000000000000000000bc0aed017c5687000000000000000000")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BInfoData(
            battery_volts=3.3,
            temperature=23,
            date_time=timedelta(seconds=567648000.0),
            relay_is_active=False,
        ),
        ),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BInfoData(
            battery_volts=3.3,
            temperature=23,
            date_time=timedelta(seconds=567648000.0),
            relay_is_active=False,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("7300cc380000000000000000000000007300cc38000000000000000000000000")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BJournalData(
            days_ago=timedelta(seconds=0.0),
            valid=False,
            time_offset=timedelta(seconds=14400.0),
            journal=(
                EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
            ),
        )),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BJournalData(
            days_ago=timedelta(seconds=0.0),
            valid=False,
            time_offset=timedelta(seconds=14400.0),
            journal=(
                EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
                EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE),
            ),
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("6c060080002000060001280006e000c86c060080002000060001280006e000c8")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BProfile8H3EnergyData(
            type=SmpmUlDeviceEnergy16bProfile8h3EnergyType.ENERGY_CONSUMED_ACTIVE,
            days_ago=timedelta(seconds=0.0),
            profile=(0, 1, 2, 3, 4, 5, 6, 7),
            point_factor=5.0,
            point_factor_multiplier=2,
        ),
        ),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BProfile8H3EnergyData(
            type=SmpmUlDeviceEnergy16bProfile8h3EnergyType.ENERGY_CONSUMED_ACTIVE,
            days_ago=timedelta(seconds=0.0),
            profile=(0, 1, 2, 3, 4, 5, 6, 7),
            point_factor=5.0,
            point_factor_multiplier=2,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("6c060080002000060001280006e000c86c060080002000060001280006e000c8")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BProfile8H3EnergyData(
            type=SmpmUlDeviceEnergy16bProfile8h3EnergyType.ENERGY_CONSUMED_ACTIVE,
            days_ago=timedelta(seconds=0.0),
            profile=(0, 1, 2, 3, 4, 5, 6, 7),
            point_factor=5.0,
            point_factor_multiplier=2,
        ),
        ),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BProfile8H3EnergyData(
            type=SmpmUlDeviceEnergy16bProfile8h3EnergyType.ENERGY_CONSUMED_ACTIVE,
            days_ago=timedelta(seconds=0.0),
            profile=(0, 1, 2, 3, 4, 5, 6, 7),
            point_factor=5.0,
            point_factor_multiplier=2,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("6c060080002000060001280006e000c86c060080002000060001280006e000c8")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BProfile8H3EnergyData(
            type=SmpmUlDeviceEnergy16bProfile8h3EnergyType.ENERGY_CONSUMED_ACTIVE,
            days_ago=timedelta(seconds=0.0),
            profile=(0, 1, 2, 3, 4, 5, 6, 7),
            point_factor=5.0,
            point_factor_multiplier=2,
        ),
        ),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BProfile8H3EnergyData(
            type=SmpmUlDeviceEnergy16bProfile8h3EnergyType.ENERGY_CONSUMED_ACTIVE,
            days_ago=timedelta(seconds=0.0),
            profile=(0, 1, 2, 3, 4, 5, 6, 7),
            point_factor=5.0,
            point_factor_multiplier=2,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("c20250320f1e60873948490500000000c20250320f1e60873948490500000000")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BTariffConsumedData(
            energy_is_reactive=False,
            days_ago=timedelta(seconds=0.0),
            valid=False,
            tariff_mask=(True, False, True, False, False, True, False, False),
            slot_0=123123,
            slot_1=4312123,
            slot_2=5413,
            slot_3=0,
        ),
        ),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BTariffConsumedData(
            energy_is_reactive=False,
            days_ago=timedelta(seconds=0.0),
            valid=False,
            tariff_mask=(True, False, True, False, False, True, False, False),
            slot_0=123123,
            slot_1=4312123,
            slot_2=5413,
            slot_3=0,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_ENERGY_METER_V0.parse(bytes.fromhex("c10250320f1e60873948490500000000c10250320f1e60873948490500000000")) == [  # type: ignore
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BTariffGeneratedData(
            energy_is_reactive=False,
            days_ago=timedelta(seconds=0.0),
            valid=False,
            tariff_mask=(True, False, True, False, False, True, False, False),
            slot_0=123123,
            slot_1=4312123,
            slot_2=5413,
            slot_3=0,
        ),
        ),
        SmpMEnergyMeterV0DevicePacket(packet=SmpmUlDeviceEnergy16BTariffGeneratedData(
            energy_is_reactive=False,
            days_ago=timedelta(seconds=0.0),
            valid=False,
            tariff_mask=(True, False, True, False, False, True, False, False),
            slot_0=123123,
            slot_1=4312123,
            slot_2=5413,
            slot_3=0,
        ),
        ),
    ]


def test_protocol_smpm_gas_v0_multiple_parsing() -> None:
    assert DeviceProtocolType.SMP_M_GAS_METER_V0.parse(bytes.fromhex("840cc0ffff7fba90e4eab10623250a087c0db8010a000a00b61d862a0801e403820c188416011a00000000007d55030064e4eab1067b2100")) == [  # type: ignore
        SmpMGasMeterV0DevicePacket(packet=SmpmUlDeviceGazFlow32BDailyData(
            days_ago=timedelta(seconds=0.0),
            sync_time_days_ago=timedelta(seconds=0.0),
            timestamp_s=timedelta(seconds=33554431.0),
            temperature=23, battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=True,
            direct_flow_volume=112323.3,
            direct_flow_volume_day_ago=3.5,
            reverse_flow_volume=10.98,
            event_battery_warn=True,
            event_system_error=False,
            event_flow_reverse=False,
            event_flow_speed_is_over_limit=False,
            event_sensor_error=False,
            event_sensor_error_temperature=False,
            event_case_was_opened=False,
            event_continuous_consumption=False,
            event_no_resource=True,
            earfcn=3452,
            pci=440,
            cell_id=10,
            tac=10,
            rsrp=42,
            rsrq=9,
            rssi=24,
            snr=22,
            band=8,
            ecl=1,
            tx_power=100,
            operation_mode=3,
        ),
        ),

        SmpMGasMeterV0DevicePacket(packet=SmpmUlDeviceGazFlow16BDailyData(
            days_ago=timedelta(0),
            sync_time_days_ago=timedelta(days=3),
            timestamp_s=timedelta(days=3, seconds=26000),
            temperature=-9,
            battery_volts=0.0,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=False,
            direct_flow_volume=0.0,
            rsrp=-15,
            rsrq=-9,
            snr=22,
            ecl=1,
            event_battery_warn=False,
            event_system_error=False,
            event_flow_reverse=False,
            event_flow_speed_is_over_limit=False,
            event_sensor_error=False,
            event_sensor_error_temperature=False,
            event_case_was_opened=False,
            event_continuous_consumption=False,
            event_no_resource=False,
        ),
        ),
        SmpMGasMeterV0DevicePacket(packet=SmpmUlDeviceGazFlow08BDailyData(
            cumulative_volume=112323.3,
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_case_was_opened=False,
            event_flow_reverse=False,
            event_flow_speed_is_over_limit=False,
            event_sensor_error_measurement=False,
            event_sensor_error_temperature=False,
            event_low_battery_level=False,
            event_system_error=False,
        ),
        ),
    ]


def test_protocol_smpm_jupiter_multiple_parsing() -> None:
    assert DeviceProtocolType.SMP_M_JUPITER_08B_V0.parse(bytes.fromhex("65fffffffbffffdf66ffffff7fa16e0067ffffff7fa16e00")) == [  # type: ignore
        SmpMJupiter08BV0DevicePacket(packet=SmpmUlDeviceJupiter08BCounterData(
            value_channel_1=67108863,
            value_channel_2=67108863,
            event_reset=True,
            event_low_battery_level=True,
        ),
        ),
        SmpMJupiter08BV0DevicePacket(packet=SmpmUlDeviceJupiter08BCounterCh1Data(
            value=2147483647,
            battery_volts=3.3,
            temperature=23,
            event_reset=True,
            event_low_battery_level=True,
        ),
        ),
        SmpMJupiter08BV0DevicePacket(packet=SmpmUlDeviceJupiter08BCounterCh2Data(
            value=2147483647,
            battery_volts=3.3,
            temperature=23,
            event_reset=True,
            event_low_battery_level=True,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_JUPITER_12B_V0.parse(bytes.fromhex("d4f9ffffffefffffffbf5037d4f9ffffffefffffffbf5037")) == [  # type: ignore
        SmpMJupiter12BV0DevicePacket(packet=SmpmUlDeviceJupiter12BCounterData(
            value_channel_1=8589934591,
            value_channel_2=8589934591,
            battery_volts=3.3,
            temperature=23,
            event_reset=True,
            event_low_battery_level=True,
        ),
        ),
        SmpMJupiter12BV0DevicePacket(packet=SmpmUlDeviceJupiter12BCounterData(
            value_channel_1=8589934591,
            value_channel_2=8589934591,
            battery_volts=3.3,
            temperature=23,
            event_reset=True,
            event_low_battery_level=True,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_JUPITER_16B_V0.parse(bytes.fromhex("810cc0ffff7fba90fdfffffffeffff7f810cc0ffff7fba90fdfffffffeffff7f")) == [  # type: ignore
        SmpMJupiter16BV0DevicePacket(packet=SmpmUlDeviceJupiter16BCounterData(
            days_ago=timedelta(seconds=0.0),
            sync_time_days_ago=timedelta(seconds=0.0),
            timestamp_s=timedelta(seconds=33554431.0),
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=True,
            event_battery_warn=True,
            event_system_error=False,
            value_channel_1=1073741823,
            value_channel_2=1073741823,
        ),
        ),
        SmpMJupiter16BV0DevicePacket(packet=SmpmUlDeviceJupiter16BCounterData(
            days_ago=timedelta(seconds=0.0),
            sync_time_days_ago=timedelta(seconds=0.0),
            timestamp_s=timedelta(seconds=33554431.0),
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=True,
            event_battery_warn=True,
            event_system_error=False,
            value_channel_1=1073741823,
            value_channel_2=1073741823,
        ),
        ),
    ]


def test_protocol_smpm_water_meter_multiple_parsing() -> None:
    assert DeviceProtocolType.SMP_M_WATER_METER_04B_V0.parse(bytes.fromhex('0ce13a060ce13a06')) == [  # type: ignore
        SmpMWaterMeter04BV0DevicePacket(packet=SmpmUlDeviceWaterMeter04BEventData(
            battery_volts=3.3,
            event_low_battery_level=True,
            event_temperature_limits=True,
            temperature=23,
            event_case_was_opened=False,
            event_magnet=False,
            event_reset=True,
            event_sensor_error=True,
        ),
        ),
        SmpMWaterMeter04BV0DevicePacket(packet=SmpmUlDeviceWaterMeter04BEventData(
            battery_volts=3.3,
            event_low_battery_level=True,
            event_temperature_limits=True,
            temperature=23,
            event_case_was_opened=False,
            event_magnet=False,
            event_reset=True,
            event_sensor_error=True,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_WATER_METER_08B_V0.parse(bytes.fromhex('68e4eab106a33a7874e4eab1d685cf4069e14a1400000000')) == [  # type: ignore
        SmpMWaterMeter08BV0DevicePacket(packet=SmpmUlDeviceWaterMeter08BDailyData(
            direct_flow_volume=112323.3,
            direct_flow_volume_day_ago=3.5,
            event_battery_or_temperature_limits=True,
            temperature=23,
            event_magnet=False,
            event_continues_consumption=False,
            event_case_was_opened=False,
            event_system_error=False,
            event_reset=True,
            event_sensor_error=True,
            event_no_resource=True,
            event_flow_speed_is_over_limit=True,
            event_flow_reverse=False,
        ),
        ),
        SmpMWaterMeter08BV0DevicePacket(packet=SmpmUlDeviceWaterMeter08BDailyNewData(
            value=112323.3,
            temperature=23,
            battery_volts=3.3,
            event_low_battery_level=True,
            event_temperature_limits=True,
            event_reset=True,
            event_magnet=True,
            event_reverse_flow=False,
            value_reverse=0.518,
        ),
        ),
        SmpMWaterMeter08BV0DevicePacket(packet=SmpmUlDeviceWaterMeter08BInfoData(
            battery_volts=3.3,
            event_battery_low=True,
            event_battery_warn=True,
            reverse_flow_volume=10.98,
            event_flow_reverse=True,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_WATER_METER_12B_V0.parse(bytes.fromhex('63c0e4eab1062337bae14a2463c0e4eab1062337bae14a24')) == [  # type: ignore
        SmpMWaterMeter12BV0DevicePacket(packet=SmpmUlDeviceWaterMeter12BDailyData(
            days_ago=timedelta(seconds=0.0),
            event_case_was_opened=False,
            event_reset=True,
            event_sensor_error=True,
            direct_flow_volume=112323.3,
            direct_flow_volume_day_ago=3.5,
            event_magnet=False,
            direct_flow_volume_2days_ago=5.5,
            event_continues_consumption=False,
            temperature=23,
            event_temperature_limits=True,
            battery_volts=3.3,
            event_battery_warn=True,
            event_flow_speed_is_over_limit=True,
            reverse_flow_volume=1.098,
            event_flow_reverse=False,
            event_no_resource=True,
            event_system_error=False,
        ),
        ),
        SmpMWaterMeter12BV0DevicePacket(packet=SmpmUlDeviceWaterMeter12BDailyData(
            days_ago=timedelta(seconds=0.0),
            event_case_was_opened=False,
            event_reset=True,
            event_sensor_error=True,
            direct_flow_volume=112323.3,
            direct_flow_volume_day_ago=3.5,
            event_magnet=False,
            direct_flow_volume_2days_ago=5.5,
            event_continues_consumption=False,
            temperature=23,
            event_temperature_limits=True,
            battery_volts=3.3,
            event_battery_warn=True,
            event_flow_speed_is_over_limit=True,
            reverse_flow_volume=1.098,
            event_flow_reverse=False,
            event_no_resource=True,
            event_system_error=False,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_WATER_METER_16B_V0.parse(bytes.fromhex('830cc0ffff7fba90e4eab10623250a08830cc0ffff7fba90e4eab10623250a08')) == [  # type: ignore
        SmpMWaterMeter16BV0DevicePacket(packet=SmpmUlDeviceWaterMeter16BDailyData(
            days_ago=timedelta(seconds=0.0),
            sync_time_days_ago=timedelta(seconds=0.0),
            timestamp_s=timedelta(seconds=33554431.0),
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=True,
            direct_flow_volume=112323.3,
            direct_flow_volume_day_ago=3.5,
            reverse_flow_volume=1.098,
            event_battery_warn=True,
            event_system_error=False,
            event_flow_reverse=False,
            event_flow_speed_is_over_limit=False,
            event_sensor_error=False,
            event_sensor_error_temperature=False,
            event_case_was_opened=False,
            event_continuous_consumption=False,
            event_no_resource=True,
            event_magnet=False,
        ),
        ),
        SmpMWaterMeter16BV0DevicePacket(packet=SmpmUlDeviceWaterMeter16BDailyData(
            days_ago=timedelta(seconds=0.0),
            sync_time_days_ago=timedelta(seconds=0.0),
            timestamp_s=timedelta(seconds=33554431.0),
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=True,
            direct_flow_volume=112323.3,
            direct_flow_volume_day_ago=3.5,
            reverse_flow_volume=1.098,
            event_battery_warn=True,
            event_system_error=False,
            event_flow_reverse=False,
            event_flow_speed_is_over_limit=False,
            event_sensor_error=False,
            event_sensor_error_temperature=False,
            event_case_was_opened=False,
            event_continuous_consumption=False,
            event_no_resource=True,
            event_magnet=False,
        ),
        ),
    ]

    assert DeviceProtocolType.SMP_M_WATER_METER_16B_V0.parse(bytes.fromhex('850cc0ffff7fba90e4eab106e4aaaf6a85cc20a09407cdd0c4c91200c4b9af6a')) == [  # type: ignore
        SmpMWaterMeter16BV0DevicePacket(packet=SmpmUlDeviceWaterMeterNbiot16BDailyData(
            days_ago=timedelta(seconds=0.0),
            sync_time_days_ago=timedelta(seconds=0.0),
            timestamp_s=timedelta(seconds=33554431.0),
            temperature=23,
            battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=False,
            event_temperature_limits=True,
            direct_flow_volume=112323.3,
            reverse_flow_volume=2.788,
            event_magnet=False,
            rsrp=-15,
            rsrq=-9,
            snr=22,
            ecl=1,
        ),
        ),
        SmpMWaterMeter16BV0DevicePacket(packet=SmpmUlDeviceWaterMeterNbiot16BDailyData(
            days_ago=timedelta(days=3),
            sync_time_days_ago=timedelta(days=4),
            timestamp_s=timedelta(days=23),
            temperature=42,
            battery_volts=3.3,
            event_reset=False,
            event_low_battery_level=True,
            event_temperature_limits=True,
            direct_flow_volume=1231.3,
            reverse_flow_volume=2.5,
            event_magnet=True,
            rsrp=-15,
            rsrq=-9,
            snr=22,
            ecl=1,
        ),
        ),
    ]


def test_protocol_water5_v0_multiple_parsing() -> None:
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('54c200000000000054c2000000000000')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceDailyData(value=24.874)),
        Water5NeroV0DevicePacket(packet=Water5DeviceDailyData(value=24.874)),
    ]

    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('73086e0a000000a573086e0a000000a5')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=683.528,
            battery_int=3,
            battery_fract=37,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=683.528,
            battery_int=3,
            battery_fract=37,
        ),
        ),
    ]

    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('7b3a00000000001f7b3a00000000001f735ABD31070000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh2Data(
            value=0.058,
            battery_int=2,
            battery_fract=31,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh2Data(
            value=0.058,
            battery_int=2,
            battery_fract=31,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000 ПОТОМУ ЧТО м3 а не в литрах
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]

    # FLUO A, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3735ABD31070000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000 ПОТОМУ ЧТО м3 а не в литрах
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000 ПОТОМУ ЧТО м3 а не в литрах
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]

    # FLUO A, W5, INFO
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC4894470000D961CC4894470000D9')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда, В ДОКЕ УКАЗАНО 00 !!!!!(возможно опечатка)
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда, В ДОКЕ УКАЗАНО 00 !!!!!(возможно опечатка)
        ),
        ),
    ]

    # FLUO A, W5, RESET
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('41CC48944701410041CC489447014100')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceColdResetData(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceColdResetData(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
    ]

    # FLUO(FREESCALE), W5, DAILY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('60860000000000006086000000000000')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceDailyData(
            value=17200 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceDailyData(
            value=17200 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
    ]

    # FLUO(FREESCALE), W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3735ABD31070000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]

    # FLUO(FREESCALE), W5, WEEKLY (исп.0), Размерность переменной накопленного объема в прямом направлении за все время равна 24 бит вместо 27 бит.
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('73DA45A3000000C373DA45A3000000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=10700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=10700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),  # counter_bit_size исходя из описания возможно максимум 24???
        ),
    ]
    # WEEKLY и WEEKLY (исп.0) дают идентичный результат

    # FLUO(FREESCALE), W5, INFO, Размерность переменной накопленного объема в прямом направлении за все время равна 24 бит вместо 27 бит.
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC4894470000D961CC4894470000D9')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
    ]

    # FLUO(FREESCALE), W5, INFO(исп.0), Присутствует значение переменной общего количества отправленных пакетов count_send_message
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC489447F875D961CC489447F875D9')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
    ]

    # FLUO(FREESCALE), W5, RESET
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('41CC48944701410041CC489447014100')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceColdResetData(
            value=1200900300 / 1000,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceColdResetData(
            value=1200900300 / 1000,
        ),
        ),
    ]

    # FLUO(FREESCALE), W5, COMMAND
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('51AAAAAAAAAAAAAA51AAAAAAAAAAAAAA')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceMagnetData()),
        Water5NeroV0DevicePacket(packet=Water5DeviceMagnetData()),
    ]

    # JUPITER FREESCALE/STM32, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31000000C3735ABD31000000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=3259738 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=3259738 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]

    # FLUO LoRa, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3735ABD31070000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]

    # FLUO STM, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3735ABD31070000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]
    # FLUO LoRa, W5, WEEKLY = FLUO STM, W5, WEEKLY

    # FLUO STM, W5, INFO
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('61CC4894470000D961CC4894470000D9')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceInfoCh1Data(
            value=1200900300 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
        ),
        ),
    ]

    # FLUO STM, W5, COMMAND
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('51AAAAAAAAAAAAAA51AAAAAAAAAAAAAA')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceMagnetData()),
        Water5NeroV0DevicePacket(packet=Water5DeviceMagnetData()),
    ]

    # Jupiter LoRa, W5, WEEKLY
    assert DeviceProtocolType.WATER5_V_NERO_V0.parse(bytes.fromhex('735ABD31070000C3735ABD31070000C3')) == [  # type: ignore
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
        Water5NeroV0DevicePacket(packet=Water5DeviceWeeklyCh1Data(
            value=120700250 / 1000,  # не совпадает поле, множитель 1000, выяснить откуда
            battery_int=3,
            battery_fract=67,
        ),
        ),
    ]



from enum import Enum, unique
from typing import Optional

from data_gateway_sdk.crypto_algorithms.crypto_aes import aes_ecb_encryption, aes_ecb_decryption
from data_gateway_sdk.crypto_algorithms.crypto_kuznechik import kuznechik_encryption, kuznechik_decryption
from data_gateway_sdk.crypto_algorithms.crypto_xtea import xtea_decrypt, xtea_encrypt


@unique
class EncryptionType(Enum):  # DO NOT CHANGE VALUE because it should be compatible with data_aggregator_sdk db !!!
    NO_ENCRYPTION = 'NO_ENCRYPTION'
    XTEA_V_NERO_V0 = 'XTEA_V_NERO_V0'
    AES_ECB_V_NERO_V0 = 'AES_ECB_V_NERO_V0'
    KUZNECHIK_V_NERO_V0 = 'KUZNECHIK_V_NERO_V0'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'

    def encrypt(self, src: bytes, key: Optional[bytes]) -> bytes:
        assert isinstance(src, bytes), f'src decrypt must be type of {bytes}, got type {type(src)} '

        if self is EncryptionType.NO_ENCRYPTION:
            assert key is None
            return src

        assert isinstance(key, bytes), f'key decrypt must be type of {bytes}, got type {type(key)} '

        if self is EncryptionType.XTEA_V_NERO_V0:
            return xtea_encrypt(src, key)

        if self is EncryptionType.AES_ECB_V_NERO_V0:
            return aes_ecb_encryption(src, key)

        if self is EncryptionType.KUZNECHIK_V_NERO_V0:
            return kuznechik_encryption(src, key)

        raise NotImplementedError(f'Unsupported type. {self} was given')

    def decrypt(self, src: bytes, key: Optional[bytes]) -> bytes:
        assert isinstance(src, bytes), f'src decrypt must be type of {bytes}, got type {type(src)} '

        if self is EncryptionType.NO_ENCRYPTION:
            assert not key, f'key decrypt must be empty on {EncryptionType.NO_ENCRYPTION}, got type {type(key)} '
            return src

        assert isinstance(key, bytes), f'key decrypt must be type of {bytes}, got type {type(key)} '

        if self is EncryptionType.XTEA_V_NERO_V0:
            return xtea_decrypt(src, key)

        if self is EncryptionType.AES_ECB_V_NERO_V0:
            return aes_ecb_decryption(src, key)

        if self is EncryptionType.KUZNECHIK_V_NERO_V0:
            return kuznechik_decryption(src, key)

        raise NotImplementedError(f'Unsupported type. {self} was given')



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_8b_set_clock import SmpmDlDeviceEnergy8BSetClockData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_dl_device_energy_8b_set_clock() -> None:
    case_serialized = bytes.fromhex("02ffffff7fffff00")
    assert SmpmDlDeviceEnergy8BSetClockData(time=timedelta(seconds=2147483647.0), time_zone_offset_s=timedelta(seconds=65535.0), time_zone_offset_is_negative=False) == SmpmDlDeviceEnergy8BSetClockData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("0200000000000002")  # overflow
    assert SmpmDlDeviceEnergy8BSetClockData(time=timedelta(seconds=0.0), time_zone_offset_s=timedelta(seconds=0.0), time_zone_offset_is_negative=True) == SmpmDlDeviceEnergy8BSetClockData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("0200000000ffff03")  # overflow
    assert SmpmDlDeviceEnergy8BSetClockData(time=timedelta(seconds=0.0), time_zone_offset_s=timedelta(seconds=131071.0), time_zone_offset_is_negative=True) == SmpmDlDeviceEnergy8BSetClockData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("02ffffffff000002")  # overflow
    assert SmpmDlDeviceEnergy8BSetClockData(time=timedelta(seconds=4294967295.0), time_zone_offset_s=timedelta(seconds=0.0), time_zone_offset_is_negative=True) == SmpmDlDeviceEnergy8BSetClockData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("02ffffffffffff03")  # overflow
    assert SmpmDlDeviceEnergy8BSetClockData(time=timedelta(seconds=4294967295.0), time_zone_offset_s=timedelta(seconds=131071.0), time_zone_offset_is_negative=True) == SmpmDlDeviceEnergy8BSetClockData.parse(BufRef(case_serialized))  # noqa: E501



from data_gateway_sdk.protocols.water5_v_nero.water5_device_info_ch2 import Water5DeviceInfoCh2Data
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_info_ch2() -> None:
    case = Water5DeviceInfoCh2Data(value=16.3)  # noqa: E501
    assert case == Water5DeviceInfoCh2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceInfoCh2Data(value=0.0)  # noqa: E501
    assert case == Water5DeviceInfoCh2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceInfoCh2Data(value=16.3)  # noqa: E501
    assert case == Water5DeviceInfoCh2Data.parse(BufRef(case.serialize()))
    case = Water5DeviceInfoCh2Data(value=4294967.295)  # noqa: E501
    assert case == Water5DeviceInfoCh2Data.parse(BufRef(case.serialize()))



from typing import Dict, List, Any
from datetime import datetime, tzinfo

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent
from data_aggregator_sdk.integration_message import IntegrationV0MessageData

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   water5_device_magnet
#
# RESULT int:        81
# RESULT bin:  MSB   0000000000000000000000000000000000000000000000000000000001010001   LSB
# RESULT hex:  LE    51 00 00 00 00 00 00 00
#
#
# name      type  size  value(int)                                                        data(bits)
# --------------------------------------------------------------------------------------------------
# pack_id   u8       8          81                                                          01010001
# RESERVED  u56     56           0  00000000000000000000000000000000000000000000000000000000


class Water5DeviceMagnetData(Packet):
    def serialize(self) -> bytes:
        result = 0
        size = 0
        result |= ((81) & (2 ** (8) - 1)) << size
        size += 8
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'Water5DeviceMagnetData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 81 != buf.shift(8):
            raise ValueError("pack_id: buffer doesn't match value")
        buf.shift(56)
        result = Water5DeviceMagnetData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                events=[IntegrationV0MessageEvent.MAGNET_WAS_DETECTED],
            ),
        ]



from decimal import Decimal
from typing import List, Any, Dict
from datetime import datetime, tzinfo

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, CounterType, ResourceType, IntegrationV0MessageGeneration, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_ul_device_water_meter_08b_info
#
# RESULT int:        340451689
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00010100 01001010 11100001 01101001   LSB
# RESULT hex:  LE    69 e1 4a 14 00 00 00 00
#
# name                  type    size  value(int)                                                        data(bits)
# --------------------  ------  ----  ----------  ----------------------------------------------------------------
# packet_type_id.0.VAL  u7         7         105                                                           1101001
# packet_type_id.0.DFF  bool       1           0                                                          0
# battery_volts         uf6p1      6          33                                                    100001
# event_battery_low     bool       1           1                                                   1
# event_battery_warn    bool       1           1                                                  1
# reverse_flow_volume   uf12p2    12        1098                                      010001001010
# event_flow_reverse    bool       1           1                                     1
# RESERVED              u35       35           0  00000000000000000000000000000000000


class SmpmUlDeviceWaterMeter08BInfoData(Packet):
    battery_volts: float
    event_battery_low: bool
    event_battery_warn: bool
    reverse_flow_volume: float
    event_flow_reverse: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((105) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_battery_low, bool)
        result |= ((int(data.event_battery_low)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_battery_warn, bool)
        result |= ((int(data.event_battery_warn)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.reverse_flow_volume, (int, float))
        result |= ((int(round(float(data.reverse_flow_volume) * 100.0, 0)) & 4095) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.event_flow_reverse, bool)
        result |= ((int(data.event_flow_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceWaterMeter08BInfoData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 105 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp1["event_battery_low"] = bool(buf.shift(1))
        result__el_tmp1["event_battery_warn"] = bool(buf.shift(1))
        result__el_tmp1["reverse_flow_volume"] = round(buf.shift(12) / 100.0, 2)
        result__el_tmp1["event_flow_reverse"] = bool(buf.shift(1))
        result = SmpmUlDeviceWaterMeter08BInfoData(**result__el_tmp1)
        buf.shift(35)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                ],
                generation=[
                    IntegrationV0MessageGeneration(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        value=Decimal(str(self.reverse_flow_volume)),
                        overloading_value=Decimal(str(40.95)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.FLOW_REVERSE] if self.event_flow_reverse else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_battery_low else []),
                    *([IntegrationV0MessageEvent.BATTERY_WARNING] if self.event_battery_warn else []),
                ],
            ),
        ]
